"""
ANN class for simple Q-function approximation
Uses SGD to update the parameters.
"""

from collections import OrderedDict
import torch
import torch.nn as nn
from torch import Tensor, LongTensor, optim

class Net( nn.Module ) :

    def __init__( self , M_I , M_H , M_O , lr , momentum=None) :

        super( Net , self ).__init__()
        self.M_I = M_I
        self.M_H = M_H
        self.M_O = M_O

        hidden_layers = OrderedDict()
        tmp_M_H = [ self.M_I ] + self.M_H
        for mh_i in range(1,len(tmp_M_H)) :
            hidden_layers[ 'layer_h'+str(mh_i) ] = nn.Linear( tmp_M_H[mh_i-1] , tmp_M_H[mh_i] )
            hidden_layers[ 'tanh'+str(mh_i) ] = nn.Tanh()
        self.hidden_layers = nn.Sequential( hidden_layers )
        self.output = nn.Linear( self.M_H[-1] , self.M_O )

        if momentum :
            self._optimizer = optim.SGD( self.parameters() , lr=lr , momentum=momentum )
        else :
            self._optimizer = optim.SGD( self.parameters() , lr=lr )
            
        self._criterion = nn.MSELoss()

    def forward( self , x , grad_p ) :

        x = torch.tensor( x , dtype=torch.float , requires_grad=False )
        with torch.set_grad_enabled( grad_p ):
            tmp = self.hidden_layers( x )
            Q = self.output( tmp )
        return Q

    def compute_Q( self , x ) :

        Q = self.forward( x , False )
        return Q.numpy()

    def update( self , X , T ) :

        self._optimizer.zero_grad()
        X_torch = torch.tensor( X , dtype=torch.float )
        T_torch = torch.tensor( T , dtype=torch.float )
        Y = self.forward( X , True)
        loss = self._criterion( Y , T_torch )
        loss.backward()
        self._optimizer.step()



# TODO: remove num_updates parameter from learn method?
# TODO: add ability to send a list of parameters to the Q-func approximator via the learn method

import random, tempfile, uuid
import numpy as np

class Qlearner:
    """
    The plan is to make this class increasingly abstract as we add more varieties of Q-learning
    """

    def __init__( self , Q_approx , actions , M_I_state , gamma=0.9):
        
        self._replay_memory = []
        self._actions = actions

        self._gamma = gamma

        self._M_I_state = M_I_state
        self._M_I_actions = len(self._actions[0])
        self._M_I = self._M_I_state + self._M_I_actions

        self._Q_approx = Q_approx

    def select_action( self , state_in , epsilon=None ):

        if not epsilon:
            epsilon = 0.0
        if random.random() > epsilon:

            # compute q-value for each action
            qvals = np.zeros( len(self._actions) )
            net_in = np.append( state_in , np.zeros( ( 1 , self._M_I_actions ) ) , axis=1 ) # add elements for action
            for ai,a in enumerate( self._actions ):
                net_in[ 0, self._M_I_state: ] = self._actions[ ai ]
                Qout = self._Q_approx.compute_Q( net_in )
                # store Q-value
                qvals[ai] = Qout[0,0]
            # select the winning action
            selected_action_i = np.argmax( qvals )

        else:
            selected_action_i = random.sample( range(len(self._actions)) , k=1 )[0]
            
        return self._actions[ selected_action_i ] , selected_action_i

    def generate_episode( self , episode_len , step_f , init_f , epsilon=None ):

        episode = []

        # get initial state for episode
        current_state = init_f()

        # loop over episode time steps
        for episode_t in range(episode_len):

            # select action
            current_action , current_action_idx = self.select_action( current_state , epsilon=epsilon )
            # advance simulation
            (next_state , current_reward , finished_p) = step_f( current_action_idx )
            # add to episode memory
            episode += [ {"s":current_state , "ai":current_action_idx , "r":current_reward , "snext":next_state } ]
            # advance to the next state
            current_state = next_state
            # exit if episode is finished
            if finished_p:
                break

        return episode

    def add_to_memory( self , episode ):
        # add to memory
        self._replay_memory += episode

    def build_training_data( self , batch_size ):
        """
        select memories
        put states into a training input numpy ndarray
        compute target values and store in a ndarray
        return both ndarrays
        """
        # 
        # input data
        #
        batch_size = min( batch_size , len( self._replay_memory ) )
        input_data = np.zeros( (batch_size , self._M_I ) )
        # store the states in the inputs
        selected_memories = random.sample( self._replay_memory , batch_size )
        input_data[ : , :self._M_I_state ] = np.concatenate( [ sm["s"] for sm in selected_memories ] , axis=0 )
        # actions make up the rest of the inputs
        input_data[ : , self._M_I_state: ] = np.stack( [ self._actions[ sm["ai"] ] for sm in selected_memories ] , axis=0 )
        
        # 
        # compute the target values for each selected memory
        # start with computing max Q(snext,a) for all actions
        #

        # DLE: shouldn't this be the next state?
        # input_data_Qnext = np.copy( input_data )

        input_data_Qnext = np.zeros_like( input_data )
        input_data_Qnext[ : , :self._M_I_state ] = np.concatenate( [ sm["snext"] for sm in selected_memories ] , axis=0 )
        Qvals_next = np.zeros( ( batch_size , len(self._actions) ) )
        for ai in range( len(self._actions) ):
            input_data_Qnext[ : , self._M_I_state: ] = np.repeat( np.expand_dims( np.array( self._actions[ai] ) , axis=0 ) , repeats=batch_size , axis=0 )
            Qvals_next[ : , ai:ai+1 ] = self._Q_approx.compute_Q( input_data_Qnext )

        max_Qvals_next = np.expand_dims( np.max( Qvals_next , axis = 1 ) , axis = 1 )

        current_rs = np.expand_dims( np.array( [ sm["r"] for sm in selected_memories ] ) , axis = 1 )

        T = current_rs + self._gamma * max_Qvals_next

        #
        # done.  return input and target
        # 
        return input_data , T

        
    def learn( self , num_updates , batch_size ):

        for update_i in range(num_updates):
            #
            # select experiences from memory to update ANN
            # 
            X , T = self.build_training_data( batch_size )

            #
            # perform parameter update
            #
            self._Q_approx.update( X , T )



import sys, math
import torch
from torch import Tensor
from torch.optim import Optimizer

floatPrecision = sys.float_info.epsilon
sigma0 = 1.0e-6
betamin = 1.0e-15
betamax = 1.0e20

def unpack(Veced,unVeced):
    startIdx = 0; stopIdx = 0;
    for oP in unVeced:
        stopIdx = startIdx + torch.numel(oP)
        # print( netp )
        oP.copy_(Veced[startIdx:stopIdx].view(*oP.size()))
        # print( netp )
        startIdx = stopIdx

def pack(unVeced):
    return torch.cat([oP.view(-1) for oP in unVeced])
    
class SCG(Optimizer):
    r"""Implements stochastic gradient descent (optionally with momentum).

    If you want to resatrt SCG, just create a new SCG instance.



    Args:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float): learning rate
        momentum (float, optional): momentum factor (default: 0)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        dampening (float, optional): dampening for momentum (default: 0)
        nesterov (bool, optional): enables Nesterov momentum (default: False)

    Example:
        >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.1, momentum=0.9)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    __ http://www.cs.toronto.edu/%7Ehinton/absps/momentum.pdf

    .. note::
        The implementation of SGD with Momentum/Nesterov subtly differs from
        Sutskever et. al. and implementations in some other frameworks.

        Considering the specific case of Momentum, the update can be written as

        .. math::
                  v = \rho * v + g \\
                  p = p - lr * v

        where p, g, v and :math:`\rho` denote the parameters, gradient, velocity, and
        momentum respectively.

        This is in constrast to Sutskever et. al. and
        other frameworks which employ an update of the form

        .. math::
             v = \rho * v + lr * g \\
             p = p - v

        The Nesterov version is analogously modified.
    """

    def __init__(self, params, verbose=False):
        defaults = dict()
        super(SCG, self).__init__(params, defaults)

        self.verbose = verbose
        self.numParams = int(torch.sum(Tensor([torch.numel(par.data) for
                                               par in self.getOptParams()])))
        self.success = True
        self.nsuccess = 0
        self.beta = 1.0e-6
        self.fold = None
        self.fnow = None
        self.gradnew = None
        self.gradold = None
        self.d = None
        self.theta = 0          # will be computed first call to step()
        self.kappa = 0          # will be computed first call to step()


    def __setstate__(self, state):
        super(SCG, self).__setstate__(state)

    def getOptParams(self):
        optParams = []
        for group in self.param_groups:
            for p in group['params']:
                optParams += [p]
        return optParams

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, not optional): A closure that reevaluates the model
                and returns the loss.
        """

        if self.d is None:
            self.fold = closure(grad=True)
            self.fnow = self.fold
            self.gradnew = pack([op.grad.data for op in self.getOptParams()])
            self.gradold = self.gradnew.clone()
            self.d = torch.mul(self.gradold , -1.)
            self.success = True
            self.nsuccess = 0
            self.beta = 1.0e-6

        # vectorize parameters and gradients
        x = pack([op.data for op in self.getOptParams()])
        #??? self.gradnew = pack([op.grad.data for op in self.getOptParams()])

        if self.success:
            self.mu = torch.dot(self.d , self.gradnew)
            if math.isnan(self.mu):
                print( "bad things have happened, mu is NaN." )
            if self.mu >= 0:
                torch.mul(self.gradnew , -1. , out=self.d)
                self.mu = torch.dot(self.d , self.gradnew)
            self.kappa = torch.dot(self.d , self.d)
            if self.kappa < floatPrecision:
                print( "reached limit on machine precision.  quitting" )
                return
            sigma = sigma0 / math.sqrt(self.kappa)
            xplus = torch.add(x , sigma , self.d)
            unpack(xplus,[op.data for op in self.getOptParams()])
            closure(grad=True)  # model holds gplus
            gplus = pack([op.grad.data for op in self.getOptParams()])
            self.theta = torch.dot(self.d , gplus.sub(self.gradnew)) / sigma

        # increase effective curvature and evaluate step size, alpha
        delta = self.theta + self.beta * self.kappa
        if math.isnan(delta): print ("delta is NaN")
        if delta <= 0:
            delta = self.beta * self.kappa
            self.beta = self.beta - self.theta/self.kappa
        # XXX: this was added by chuck?
        # if delta == 0:
        #     alpha = 1
        # else:
        alpha = -self.mu / delta
        
        ## parameter update
        xnew = torch.add(x , alpha , self.d)
        unpack(xnew , [op.data for op in self.getOptParams()])
        fnew = closure(grad=True) # !!!CLOSURE CALL HERE!!!
        ## Calculate the comparison ratio.
        Delta = 2 * (fnew - self.fold) / (alpha * self.mu)
        if not math.isnan(Delta) and Delta  >= 0:
            self.success = True
            self.nsuccess += 1
            x = xnew
            self.fnow = fnew
        else:
            self.success = False
            self.fnow = self.fold

        # if successful in taking a step, prep for a new direction
        if self.success:
            self.fold = fnew         # TODO: this line is suspect, compare to Nabney
            self.gradold = self.gradnew
            # do not need to compute new gradient,
            # model currently holds latest gradients from closure call above
            self.gradnew = pack([op.grad.data for op in self.getOptParams()])
            if torch.dot(self.gradnew , self.gradnew) == 0:
                print( "zero gradient!" )
                return self.success
        
        # record some termination-related variables to return
        # retVals = {"xPrec":alpha*self.d, "fPrecision":fnew-self.fold,
        #            "newGrad":torch.dot(self.gradnew,self.gradnew),
        #            "success":self.success}

        ## Adjust beta according to comparison ratio.
        if math.isnan(Delta) or Delta < 0.25:
            self.beta = min(4.0*self.beta, betamax)
        elif Delta > 0.75:
            self.beta = max(0.5*self.beta, betamin)

        ## Update search direction using Polak-Ribiere formula, or re-start 
        ## in direction of negative gradient after nparams steps.
        if self.nsuccess == self.numParams:
            print( "success equal to number of paramters, re-starting search direction on next call" )
            torch.mul(self.gradnew , -1 , self.d)
            self.nsuccess = 0
        elif self.success:
            # print( "success! updating search direction." )
            gamma = torch.dot(self.gradold.sub(self.gradnew) ,
                              torch.div(self.gradnew , self.mu))
            self.d.mul_(gamma)
            self.d.sub_(self.gradnew)
        else:
            pass
            # print( "no success.  back to the drawing board." )
            
        unpack(x,[op.data for op in self.getOptParams()])
        return self.success



from collections import OrderedDict
import torch
import torch.nn as nn
from torch import Tensor, LongTensor, optim

import argparse , math
import gym
import numpy as np
import Qlearners.Qlearner as ql
import Qlearners.epsilon_decay as ep_decay
import Qlearners.recipes.cartpole.cartpolebox2d as cartpole

class Net( nn.Module ) :

    def __init__( self , M_H , lr , adam_epsilon=None ,
                  adam_beta1=None , adam_beta2=None ) :

        super( Net , self ).__init__()
        self.M_I = 5
        self.M_H = M_H
        self.M_O = 1

        hidden_layers = OrderedDict()
        tmp_M_H = [ self.M_I ] + self.M_H
        for mh_i in range(1,len(tmp_M_H)) :
            hidden_layers[ 'layer_h'+str(mh_i) ] = nn.Linear( tmp_M_H[mh_i-1] , tmp_M_H[mh_i] )
            hidden_layers[ 'tanh'+str(mh_i) ] = nn.Tanh()
        self.hidden_layers = nn.Sequential( hidden_layers )
        self.output = nn.Linear( self.M_H[-1] , self.M_O )

        self._optimizer = optim.Adam( self.parameters() , lr=lr )
        self._criterion = nn.MSELoss( reduction='sum' )

    def forward( self , x , grad_p ) :

        x = torch.tensor( x , dtype=torch.float , requires_grad=False )
        with torch.set_grad_enabled( grad_p ):
            tmp = self.hidden_layers( x )
            Q = self.output( tmp )
        return Q

    def compute_Q( self , x ) :

        Q = self.forward( x , False )
        return Q.numpy()

    def update( self , X , T ) :

        X_torch = torch.tensor( X , dtype=torch.float )
        T_torch = torch.tensor( T , dtype=torch.float )
        Y = self.forward( X , True)
        loss = self._criterion( Y , T_torch )
        self._optimizer.zero_grad()
        loss.backward()
        self._optimizer.step()

# TODO: add these parameters: adam_epsilon , adam_beta1 , adam_beta2
def main( M_H , lr , gamma , numReplays , episodes_max , episode_len , batch_size , evalLength , epsilon , epsilon_decay=1.0 , epsilon_min=0.1 ) :

    actions = ( [-1.0] , [0.0] , [1.0] )

    # #
    # # prepare plotting
    # #
    # plt.ion(); firstGraph=True;

    #
    # setup some constant parameters
    #

    eval_startPos = [0]                  # start eval in middle only
    eval_startAngles = [-math.pi,0.0]    # start eval on up and down positions respectively
    eval_r_hist = [ [] for i in range( len(eval_startPos)*len(eval_startAngles) ) ]
    actionDuration = 2

    #
    # method to scale the function appoximator inputs
    # 
    def scaleState(tmpI):
        tmpI[:,0] = tmpI[:,0] / scaleState.positionRange[1]     
        tmpI[:,1] = tmpI[:,1] / scaleState.velocityRange[1]     
        tmpI[:,2] = tmpI[:,2] / scaleState.angleRange[1]
        tmpI[:,3] = tmpI[:,3] / scaleState.angleVelocityRange[1]
        return(tmpI)
    scaleState.positionRange = (-2.2,2.2)
    scaleState.velocityRange = (-6.,6.)
    scaleState.angleVelocityRange = (-14.,14.)
    scaleState.angleRange = (-math.pi,math.pi)

    #
    # method (and helper method) to generate 
    # 
    def reinforcement(angle):
        angle = abs(angle)
        if angle > math.pi * 0.75:
            return -1
        elif angle < math.pi * 0.25:
            return 1
        else:
            return 0

    # ####################
    # initialize cartpole environment
    #

    def get_domain():
        return get_domain.cpdomain
    get_domain.cpdomain = cartpole.CartPole()

    #
    # ####################

    # ####################
    # create functions needed by Qlearner
    # initialize Q-learning agent
    #

    def advance_environment_f( selected_action_idx , domain ):

        # move simulation forward, get new state, scale state, return scaled state
        for ai in range(actionDuration):
            domain.act( actions[selected_action_idx][0] )
        (x,xdot,a,adot) = domain.sense()
        # compute reward
        reward = reinforcement( a )
        # scale state
        state_vec = np.array( [ [x,xdot,a,adot] ] )
        scaleState( state_vec )
        # return scaled state and reward and indicate not done (sim always not done)
        return ( state_vec , reward , False )

    def reset_domain_f():

        # doesn't actually reset: just picks up where previous episode left off
        # if want to start from same position, need to create a new instance of simulation

        domain = get_domain()
        (x,xdot,a,adot) = domain.sense()
        state_vec = np.array( [ [x,xdot,a,adot] ] )
        scaleState( state_vec )
        return state_vec


    def reset_domain_eval_f( domain ):

        (x,xdot,a,adot) = domain.sense()
        state_vec = np.array( [ [x,xdot,a,adot] ] )
        scaleState( state_vec )
        return state_vec


    Q_approx = Net(  M_H , lr=lr )
    agent = ql.Qlearner( Q_approx , actions , 4 , gamma=gamma )

    # 
    # ####################

    # ####################
    # loop over episodes
    # 

    episode_i = 0 ; train_r_hist = [] ; eval_r_hist = [] ;
    while episode_i < episodes_max:

        print( 'epsilon='  , epsilon )
        episode = agent.generate_episode( episode_len ,
                                          step_f = lambda sa: advance_environment_f(sa , get_domain() ) ,
                                          init_f = reset_domain_f ,
                                          epsilon = 0.1 )
        epsilon = ep_decay.decay( epsilon , epsilon_decay , epsilon_min )
        agent.add_to_memory( episode )
        agent.learn( num_updates=numReplays , batch_size=batch_size )


        eval_domain = cartpole.CartPole()
        eval_episode = agent.generate_episode( evalLength ,
                                               step_f = lambda sa: advance_environment_f( sa , eval_domain ) ,
                                               init_f = lambda: reset_domain_eval_f( eval_domain ) ,
                                               epsilon = None )

        eval_sum_r = np.sum( [ ee["r"] for ee in eval_episode ] )
        train_sum_r = np.sum( [ te["r"] for te in episode ] )

        print( "episode" , episode_i , "eval reward=" , eval_sum_r )

        train_r_hist += np.sum( [ ee["r"] for ee in episode ] )
        eval_r_hist += np.sum( [ ee["r"] for ee in episode ] )

        # the episode has concluded
        episode_i += 1
    
    # 
    # ####################

    return eval_r_hist , train_r_hist


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument( "-v", action="store_true", default=False, help="not implemented")
    parser.add_argument( "--alg" , default="scg" )
    parser.add_argument( "--M_H", nargs="+", type=int, default=[20,20])
    parser.add_argument( "--lr", type=float, default=0.0001)
    parser.add_argument( "--beta1", type=float, default=0.9)
    parser.add_argument( "--beta2", type=float, default=0.999)
    parser.add_argument( "--adam_eps", type=float, default=pow(10,-8) )
    parser.add_argument( "--epsilon", type=float, default=0.1 )
    parser.add_argument( "--epsilon_decay", type=float, default=1. ) # no decay
    parser.add_argument( "--epsilon_min", type=float, default=0.1 )
    parser.add_argument( "--gamma", type=float, default=0.9)
    parser.add_argument( "--numReplays", type=int, default=5)
    parser.add_argument( "--episode_len", type=int, default=1000)
    parser.add_argument( "--episodes_max", type=int, default=200)
    parser.add_argument( "--batch_size", type=int, default=1000)
    parser.add_argument( "--graph",action="store_true", default=False)
    parser.add_argument( "--evalLength" , type=int , default=2000)
    parser.add_argument( "--saveEvalHist",action="store_true", default=False)
    parser.add_argument( "--saveDir")
    parser.add_argument( "--savePrefix")
    args = parser.parse_args()

    main( M_H=args.M_H , lr=args.lr , gamma=args.gamma ,
          numReplays=args.numReplays , episodes_max=args.episodes_max , episode_len=args.episode_len ,
          batch_size=args.batch_size , evalLength=args.evalLength ,
          epsilon=args.epsilon , epsilon_decay=args.epsilon_decay , epsilon_min=0.1 )

    if args.saveEvalHist:
        saveEvalFile = tempfile.NamedTemporaryFile(mode="w",delete=False,
                                                   dir=args.saveDir,
                                                   prefix=args.savePrefix,
                                                   suffix=".outfile")




import setuptools

exec(open('Qlearners/version.py').read())

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Qlearners",
    version=__version__,
    author="D.L. Elliott",
    author_email="danelliottster@gmail.com",
    description="A package for Q learning (and friends)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/danelliottster/Qlearners",
    packages=setuptools.find_packages(),
    classifiers=[
        "Intended Audience :: Science/Research" ,
        "Programming Language :: Python :: 3" ,
        "License :: OSI Approved :: MIT License" ,
        "Development Status :: 3 - Alpha" ,
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
    python_requires='>=3.5',
)
